import sys
import argparse

from acdc_nn import util
from acdc_nn import nn

import numpy as np
import pandas as pd

from pkg_resources import resource_filename
import functools


def load_nn(weight_path):
	num_H=[32,16]
	d=0.2
	model1, model2 = nn.ACDC(num_H,d,25)
	weight_path = resource_filename('acdc_nn', 'weights/full_dataset_TL')
	model1.load_weights(weight_path)
	return model1, model2

@functools.lru_cache(100) # memory cache to avoid reloading the same protein data multiple times
def	load_prot(profile_path, pdb_path, chain):
	r = dict(zip(['structure', 'pchain', 'seq', 'd_seq2pdb', 'd_pdb2seq'], util.pdb2info(pdb_path, chain)))
	r['prof'] = util.getProfile(profile_path)
	r['profile_path'] = profile_path
	r['pdb_path'] = pdb_path
	return r

def compute_nn_input(mut, prot):
	kvar = (mut[0], prot['d_pdb2seq'][mut[1:-1]], mut[-1]) # position in the sequence
	kvar_pdb = (mut[0], mut[1:-1], mut[-1]) # position in the pdb
	from_residue = prot['seq'][int(kvar[1]) - 1]
	if kvar[0] != from_residue:
		print('WARNING: mutation {} does not correspond to residue {} from {}'.format(kvar[0], from_residue, prot['pdb_path']))

	dist_neigh_3d = util.get_neigh_ps(kvar_pdb, 5, prot['d_seq2pdb'], prot['pchain']) 
	list_dist_neigh_3d = dist_neigh_3d[kvar]

	# extracting features
	codif = util.getMutCod(mut)
	all_profile = util.Unified_prof(kvar[1], prot['prof'], prot['seq'], list_dist_neigh_3d)
	#all_data=[*codif,*all_profile,*np.zeros(600-len(all_profile))]
	return pd.DataFrame([*codif, *all_profile, *np.zeros(600 - len(all_profile))]).T

def predict(model, wt_mutation, wt_profile, wt_structure, wt_chain, mt_mutation, mt_profile, mt_structure, mt_chain):
		wt_prot = load_prot(wt_profile, wt_structure, wt_chain)
		to_predict_dir = compute_nn_input(wt_mutation, wt_prot)
		if mt_mutation is None: # compute input features for the inverse mutation inverting the input features for the direct mutation
			to_predict_inv = to_predict_dir.copy()
			to_predict_inv.iloc[:, :20] = to_predict_inv.iloc[:, :20].replace([1.0, -1.0],[-1.0, 1.0])
		else: # compute input features for the inverse mutation using the mutated protein
			mt_prot = load_prot(mt_profile, mt_structure, mt_chain)
			to_predict_inv = compute_nn_input(mt_mutation, mt_prot)
	
		# Making input in the proper shape 
		Xm_d, X1D_d, X3D_d = nn.mkInp(np.asarray(to_predict_dir).astype(np.float32), 500)
		Xm_i, X1D_i, X3D_i = nn.mkInp(np.asarray(to_predict_inv).astype(np.float32), 500)

		prediction = model.predict([X3D_d, X1D_d, Xm_d , X3D_i, X1D_i, Xm_i])
		return prediction[0][0][0]
	

def main():
	parser = argparse.ArgumentParser(description='Predict the DDG of one or more mutations.')
	parser.add_argument('--output', metavar='OUTFILE', type=argparse.FileType('w'), default=sys.stdout, help='output file path, use stdout otherwise')
	parser.add_argument('--weights', metavar='NN-WEIGHTS', 
		default=resource_filename('acdc_nn', 'weights/full_dataset_TL'), 
		help='path to the network weights, in tensorflow format') # FIXME what is the name of the format?
	subparsers = parser.add_subparsers(dest='command', metavar='CMD', required=True)

	single_parser = subparsers.add_parser('single', help='predict a single mutation')
	single_parser.add_argument('wt_mutation', metavar='MUT', help='mutation in the form Q339N')
	single_parser.add_argument('wt_profile', metavar='PROF', help='')
	single_parser.add_argument('wt_structure', metavar='PDB', help='')
	single_parser.add_argument('wt_chain', metavar='CHAIN', help='')

	inverse_parser = subparsers.add_parser('inverse', help='predict a single mutation using also the inverse structure')
	inverse_parser.add_argument('wt_mutation', metavar='MUT', help='mutation in the form Q339N')
	inverse_parser.add_argument('wt_profile', metavar='WT-PROF', help='')
	inverse_parser.add_argument('wt_structure', metavar='WT-PDB', help='')
	inverse_parser.add_argument('wt_chain', metavar='WT-CHAIN', help='')
	inverse_parser.add_argument('mt_mutation', metavar='INV-MUT', help='inverse mutation in the form Q339N')
	inverse_parser.add_argument('mt_profile', metavar='MT-PROF', help='')
	inverse_parser.add_argument('mt_structure', metavar='MT-PDB', help='')
	inverse_parser.add_argument('mt_chain', metavar='MT-CHAIN', help='')

	batch_parser = subparsers.add_parser('batch', help='predict a list of mutations from a file')
	batch_parser.add_argument('mutation_table', metavar='MUT-FILE', type=argparse.FileType('r'), help='mutation table')

	#profile_parser = subparsers.add_parser('profile', help='compute the profile from multiple alignment file') # we could also simply accept a multialn file and make a profile
	#profile_parser.add_argument('multialn', metavar='ALN-FILE', type=argparse.FileType('r'))

	args = parser.parse_args()
	base_arg_list = ['wt_mutation', 'wt_profile', 'wt_structure', 'wt_chain', 'mt_mutation', 'mt_profile', 'mt_structure', 'mt_chain']

	model1, model2 = load_nn(args.weights) # needed by all command
	if args.command in {'single', 'inverse'}:
		args_dict = vars(args)
		mutation_args = [args_dict.get(a) for a in base_arg_list]
		ddg = predict(model1, *mutation_args)
		print(ddg, file=args.output)
	elif args.command == 'batch':
		muts = pd.read_csv(args.mutation_table, sep='\s+', names=base_arg_list)
		print(muts)
		for row, mut in muts.iterrows():
			# check na values
			if mut[base_arg_list[:4]].count() < 4:
				print('ERROR at line {}: missing values in the first four columns of batch file'.format(row + 1), file=sys.stderr)
				sys.exit(1)
			if mut[base_arg_list[4:]].count() not in {0, 4}:
				print('ERROR at line {}: columns 4-8 of the batch file must be either all missing or all present'.format(row + 1), file=sys.stderr)
				sys.exit(1)
					
			mutation_args = [None if pd.isnull(v) else v for v in mut.values]
			ddg = predict(model1, *mutation_args)
			print(ddg, file=args.output)
		
	#elif args.command == 'profile':
	else:
		raise NotImplementedError('Command {} not implemented'.format(args.command))

if __name__ == '__main__':
	main()



import gzip
import numpy as np
from Bio.PDB import *
from Bio.PDB.Polypeptide import three_to_one, is_aa


def pdb2seq(pp):
    ''' pdb2seq(pp) takes a pdb_structure_chain 
    and return its sequence '''
    seq = [] # pp.get_sequence()
    reslist = []
    for ppc  in pp:
        reslist += [res for res in ppc]
        seq += [str(ppc.get_sequence())]
    return "".join(seq)

def map_pdb_pos(pp):
    ''' map_pdb_pos
    Returns two dicts seq2pdb[seq_pos], pdb2seq[pdb_pos]'''
    reslist = []
    for ppc  in pp:
        reslist += [res for res in ppc]
    seq2pdb = dict(zip( map(str,range(1,len(reslist)+1)), [str(r.get_id()[1])+r.get_id()[2].strip() for r in reslist]))
    pdb2seq = dict(zip( [str(r.get_id()[1])+r.get_id()[2].strip() for r in reslist], map(str,range(1,len(reslist)+1)) ))
    return seq2pdb, pdb2seq

def magic_open(path):
    return (gzip.open if path.endswith('.gz') else open)(path, 'rt')

def pdb2info(pdb_file, chain):
    ''' pdb2info(pdb_file) 
    Returns structure, polypeptide '''
    parser=PDBParser(QUIET=True)
    with magic_open(pdb_file) as f:
        structure = parser.get_structure('X', f)
    pchain=structure[0][chain]
    ppb=PPBuilder()
    pp = ppb.build_peptides(pchain, aa_only=False) #[0]
    return (structure, pchain, pdb2seq(pp), *map_pdb_pos(pp)) 

def mk_d_renumb_KseqVpdb(structure):
    d={}
    ppb=PPBuilder()
    s=""
    for pp in ppb.build_peptides(structure):
        s+=pp.get_sequence()
    i=1
    for residue in structure:
        t,n,lab=residue.get_id()
        if t == ' ':
            #d[(n,lab.strip())]=i
            d[str(i)]=str(n)+lab.strip()
            i+=1
    return d

def invert_keyval(din): 
    do={}
    for kin in din: do[din[kin]]=kin
    return do


def getMutCod(mut):
    code=list(np.zeros(20))
    d = {'A': 0, 'C': 1, 'D': 2,'E':3,'F':4,'G':5,'H':6,'I':7,'K':8,'L':9,
      'M':10,'N':11,'P':12,'Q':13,'R':14,'S':15,'T':16,'V':17,'W':18,'Y':19}
    w=mut[0]
    m=mut[-1]
    pos_w=d[w]
    pos_m=d[m]
    cod_mut=code.copy()
    cod_mut[pos_w]=-1
    cod_mut[pos_m]=1
    return cod_mut


def getProfile(profile_path):
    prof = {}
    lines = magic_open(profile_path).readlines()
    # POS A C D E F G H I K L M N P Q R S T V W Y -
    #aa = lines[0].split()[1:-1]
    aa = lines[0].split()[1:]
    for line in lines[1:]:
         v = line.split()
         prof[int(v[0])]= dict(zip(aa, map(float,v[1:])))
    return prof


def get_neighb(pchain,pos,radius_angst):
    # Next line: A for atoms 
    all_atom_list = Selection.unfold_entities(pchain, 'A')
    ns = NeighborSearch(all_atom_list)
    dv, dist={}, {} 
    dv[pos]=[]
    dist[pos] = []
    try: p,rlabel=int(pos),' '
    except: p,rlabel=int(pos[:-1]),pos[-1]
    for atomo in pchain[(' ',p,rlabel)]: # for atomo in chain[p]:
       rvicini = ns.search(atomo.get_coord(), radius_angst, "R")
       for aa in rvicini:
            l_dist = []
            if (aa.get_id()[0]==' ') and (aa.get_id()[1]!=p):
                for atoms in aa:
                    l_dist.append(atomo-atoms)
                if aa not in dv[pos]: 
                    dv[pos].append(aa)  # Returns object
                    dist[pos].append((aa, min(l_dist)))
                else:
                    for bb in dist[pos]:
                        if (bb[0] == aa) and (min(l_dist) < bb[1]):
                            dist[pos].remove(bb)
                            dist[pos].append((aa, min(l_dist)))
    return dv, dist
    
    

def get_neigh_pp(pchain,mut_pdb,Cangstroms):
    pdb_pos=mut_pdb[1]
    # Compute neighbor for mutated positions
    dneigh, dist =get_neighb(pchain,pdb_pos, Cangstroms)
    # Makes a dict: {mutatedpos_pdbnum : [ neighbor1, neighbor2, ... ], ...}
    dout={}
    dout[mut_pdb]=[]
    for aa in dist[pdb_pos]:
        pdbpos=(str(aa[0].get_id()[1])+aa[0].get_id()[2]).strip()
        dout[mut_pdb].append((three_to_one(aa[0].get_resname())+pdbpos, round(aa[1], 2)))
    return dout




def get_neigh_ps(mut_pdb,Cangstroms,d_seq2pdb_numb,pdbchain):
    '''From pdb to seq'''
    # Reads mutated positions for each pdb
    d_pdb=get_neigh_pp(pdbchain,mut_pdb,Cangstroms)
    dist_all={}
    d_pdb2seq_numb=invert_keyval(d_seq2pdb_numb)
    mut = list(d_pdb.keys())[0]
    mut_seqn=(mut[0],d_pdb2seq_numb[mut[1]],mut[2])
    dist_all[mut_seqn]=[]
    for wtpos in d_pdb[mut]:
        wt=wtpos[0][0]
        seqpos=d_pdb2seq_numb[wtpos[0][1:]]
        dist_all[mut_seqn].append((wt+seqpos, wtpos[1]))
    return dist_all
    

aa_modif={"TRN":"TRP", "CSO":"CYS","M3L":"LYS", "PCA":"GLU"}



if False:
    def mk_d_renumb_KseqVpdb_onlyCA(structure):
        s,d="",{}
        for res in structure:
            a3=res.get_resname()
            if a3 in aa_modif.keys():
                    pass
            else:
                  if is_aa(a3):
                       s+=three_to_one(a3)
        i=1
        for residue in structure:
            t,n,lab=residue.get_id()
            if t == ' ':
                d[str(i)]=str(n)+lab.strip()
                i+=1


    def get_drenum_seq2pdb(pdbkcode,pdbfile):
        chain=pdbkcode[-1]
        parser=PDBParser(QUIET=True)
        structure=parser.get_structure('X',pdbfile)
        if pdbkcode[:-1].lower()=='1lrp': dpos=mk_d_renumb_KseqVpdb_onlyCA(structure[0][chain])
        else: dpos=mk_d_renumb_KseqVpdb(structure[0][chain])
        return dpos


def Unified_prof(mut,profile,seq,l_dist_3d):
    pos_mut=int(mut)
    #sequence first
    prof_seq=profseq(pos_mut,profile, seq)
    #the rest of the 3d neigh        
    l_dist_3d.sort(key = lambda elem: (elem[1], elem[0]))
    prof_3d=prof3d(pos_mut,l_dist_3d, profile)
    return(prof_seq+prof_3d)

def profseq(pos_mut,profile, seq):
    prof_s=[]
    for j in range(-2,+3):
        if (pos_mut+j)>=1 and (pos_mut+j) <=len(seq):
            prof_s.append(list(profile[pos_mut+j].values())[:-1])
        else: prof_s.append(list(np.zeros(20)))
    unified_prof_s=[x for l in prof_s for x in l]
    return unified_prof_s


def prof3d(pos_mut,l_dist_3d, profile):
    profiles=[]
    for i in l_dist_3d:
        pos_res = int(i[0][1:])
        avoid_res=np.arange(-2,3)+pos_mut
        if pos_res not in avoid_res:  profiles.append(list(profile[pos_res].values())[:-1])       
    unified_prof3d=[x for l in profiles for x in l]
    return unified_prof3d




from silence_tensorflow import silence_tensorflow
silence_tensorflow()

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.layers import Lambda, Dense, Input, Dropout, Activation, Flatten, BatchNormalization, Dropout
from tensorflow.keras import backend
from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Dot, Conv1D, concatenate, GlobalAveragePooling1D
from tensorflow.keras import backend as K


# Build the network with global average.
def ACDC(num_H, d, num_3d):
    Coding = 20 # size of the coding, 20 residues, 
    #
    num_Env3D = num_3d   # number of 3D neighbours
    num_Env3D_i = num_3d   # number of 3D neighbours
    num_Env1D = 5   # number of sequeunce neighbours
    num_Env1D_i = 5 # number of sequence neighbours
    num_IMut = Coding
    num_IMut_i = Coding
    #
    # Create an input for the 3D neighbours
    input_env3D = Input(shape=(num_Env3D,Coding), name="input_env3D") # direct 
    input_env3D_i = Input(shape=(num_Env3D_i,Coding), name="input_env_i3D") # inverse
    #
    conv_3D = Conv1D(filters=Coding, kernel_size=1, strides=1, use_bias=False, name="net_envConv3D")
    #
    net_env3D = conv_3D(input_env3D) # direct 
    net_env3D_i = conv_3D(input_env3D_i) # inverse
    # create the input for the mutation for 3D
    # direct
    inputMut = Input(shape=(num_IMut,), name="inputMut")
    pool = GlobalAveragePooling1D()(net_env3D)###
    dot_prod = Dot(axes=(1,1))([pool,inputMut]) 
    # inverse
    inputMut_i = Input(shape=(num_IMut_i,), name="inputMut_i")
    pool_i = GlobalAveragePooling1D()(net_env3D_i)###
    dot_prod_i = Dot(axes=(1,1))([pool_i,inputMut_i])
    # Create an input for the Sequence neighbours
    input_env1D = Input(shape=(num_Env1D,Coding), name="input_env1D") # direct
    input_env1D_i = Input(shape=(num_Env1D_i,Coding), name="input_env_i1D") # inverse
    #
    conv_1D = Conv1D(filters=Coding, kernel_size=1, strides=1, use_bias=False, name="net_envConv1D")
    #
    net_env1D = conv_1D(input_env1D) # direct
    net_env1D_i = conv_1D(input_env1D_i) # inverse
    #
    # create the input for the mutation for 1D
    dot_prod1D = Dot(axes=(2,1))([net_env1D,inputMut])
    dot_prod1D_i = Dot(axes=(2,1))([net_env1D_i,inputMut_i])
    #
    # flat dense layer after the convolutional
    net = Sequential(name="TopLayer")
    #num_H=[32,16] # hidden nodes for each layer starting from the bottom
    i = 1
    for hidden in num_H:
        net.add(Dense(hidden, activation="relu", name="Dense_"+str(i)))
        net.add(Dropout(d)) #0.2
        i+=1
    net.add(Dense(1, activation='linear', name="Dense_"+str(i)))
    # Concatenate 3D and 1D inputs
    # direct
    conc_dir = concatenate([dot_prod,dot_prod1D, inputMut]) ###
    net_dir = net(conc_dir)
    # inverse
    conc_inv = concatenate([dot_prod_i,dot_prod1D_i, inputMut_i])###
    net_inv = net(conc_inv)
    # difference and average layers
    out_d = Lambda(diff, output_shape=diff_shape, name="out_diff")([net_dir, net_inv])
    out_av = Lambda(avout, output_shape=avout_shape, name="out_av")([net_dir, net_inv])
    # final training model
    net_dir_inv = Model([input_env3D, input_env1D, inputMut, input_env3D_i, input_env1D_i, inputMut_i],[out_d, out_av])
    # partial model, to use only one flow either direct or inverse
    net_DDG = Model([input_env3D, input_env1D, inputMut],[net_dir])
    net_dir_inv.compile(optimizer='adam',
              loss=['logcosh', abs_loss],
              loss_weights=[1., 1.] )
    return net_dir_inv, net_DDG



def diff(vects):
    x, y = vects
    return (x-y)/2.0


def diff_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)

def avout(vects):
    x, y = vects
    return (x+y)/2.0


def avout_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)
    

def abs_loss(y_true, y_pred):
    return K.mean(K.abs(y_pred), axis=-1)



def mkInp(X, col_3d):
    Xm = X[:, 0:20]
    X1D = X[:, 20:120].reshape(X.shape[0], 100//20, 20)
    X3D = X[:, 120:].reshape(X.shape[0], col_3d//20, 20)
    return Xm, X1D, X3D    



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

setuptools.setup(
	name="acdc-nn", 
	version="0.0.5",
	author="Giovanni Birolo",
	author_email="giovanni.birolo@unito.it", #FIXME
	description="A small example package", #FIXME
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/compbiomed-unito/acdc-nn",
	packages=setuptools.find_packages(),
	python_requires='>=3.6',
	install_requires=[
		'numpy>=1.19.5',
		'pandas>=1.1.5',
		'Biopython>=1.78',
		'tensorflow>=2.3.1',
		'silence_tensorflow>=1.1.1',
	],
	package_data={
		'acdc_nn': ['weights/*']
		#'tests': ['*'],
		#'': ['weights/*', 'tests/*'],
	},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	entry_points={
		'console_scripts': [
			'acdc-nn=acdc_nn.cmd:main',
		],
	}
)

